// There are 2 bugs in this program
// The first causes a complete breakdown of data coming through the data bus. 
// In the lcd_data and lcd_cursor I use a bit-roll (x>>4).
// If I have to have the data bus on the LCD hooked up to the first 4 pins on the MCU pin bank and they must be in order.
// I had them backwards.
// I first fixed this buy putting them in the right order.
// Later I rewrote the lcd_clear and lcd_data code to accept any order or choice within a pin bank.

// The second bug was in the timing.
// I polled a for loop hundreds of times a second to check for button presses during the countdown.
// 5000 loops at 1 ms per loop would be 5 seconds. Right?... Wrong!
// each loop contains lcd function I wrote that have their own delays within them.
// So I had to take those delays into account and rework the math to get the timing right.






// Launch controller
// coded by Phil Tirino

// 28 pin AVR Atmega328P
// 20x4 lcd screen on D bank
// PB3 PB4 PB5 SPI interface
// PD0 through PD6 used for LCD interface 
// PC0 through PC1 used for buttons

// Naming conventions
// All caps: register names, pin names and button names
// lower case underscore verbs: functions
// first word lower, cap second words; variables

#include <avr/io.h>                 // maps and defines names for SFR addresses       
#include <util/delay.h>				// prototypes _delay_ms and _delay_us

// pin names and port selection for LCD
// hook up RS to PD6, RW to PD5, EN to PD4, D4 to PD3, D5 to PD2, D6 to PD1, D7 to PD0 

// hook up relay to PB0

#define LCD_DDR 	DDRD			// defines data direction register for LCD
#define LCD_PORT 	PORTD			// defines port used for LCD
#define UN 			0b10000000		// unused pin, can be used as an output for other function
#define D7 			0b00000001		// 4 bit data bus
#define D6 			0b00000010		// 4 bit data bus
#define D5 			0b00000100		// 4 bit data bus
#define D4 			0b00001000		// 4 bit data bus
#define EN 			0b00010000		// the enable pin triggers the data transfer
#define RW 			0b00100000		// read write: 0 for write
#define RS 			0b01000000		// register select: 0 for instruction register and 1 for data register

// pin names for buttons

#define BUT_DDR 	DDRC			// defines data direction register for buttons
#define BUT_PORT 	PORTC			// defines port for buttons
#define BUT_PIN 	PINC			// defines pin register for buttons
#define LAUNCH     	0b00000001		// for use in bit-mask for button
#define ABORT   	0b00000010		// for use in bit-mask for button 

#define ScrollDelay1 		5

void lcd_init(void) {
	LCD_PORT |= (EN + RS + RW + D7 + D6 + D5 + D4);		// flash all pins
	_delay_ms(20);										// follow reset sequence on p. 46 HD44780U datasheet
	LCD_PORT &= UN;							 			// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4 respectively
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(5);										// wait for more than 4.1 ms
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);										// wait more than 100 us
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);										// end of reset
	LCD_PORT |= (EN + D5);								// flash 000010 to initialize 4 bit data transfer mode: sequence on p. 42 HD44780U datasheet
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D5);								// 1st nibble - flash 000010 for FUNCTION: data length 4 - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D7);								// 2nd nibble - flash 001000 for FUNCTION: display lines 2 and 5x8 font- see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for DISPLAY: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D7 + D6);							// 2nd nibble - flash 001100 for DISPLAY: display on - see p. 24
	_delay_ms(1);											// add D5 to turn cursor on and D4 to turn blinking cursor position on
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for ENTRY: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D6 + D5);							// 2nd nibble - flash 000110 for ENTRY: auto increment and shift off - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D4);								// 1st nibble - flash 000001 for CURSOR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D6);								// 2nd nibble - flash 000100 for CURSOR: cursor move shift right - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;					 					// 1st nibble - flash 000000 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D4);	 							// 2nd nibble - flash 000001 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for HOME: see p. 24
	_delay_ms(2);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D5);	 							// 2nd nibble - flash 000010 for HOME: see p. 24
	_delay_ms(2);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
}

void lcd_clear(void){
	LCD_PORT |= EN;					 					// 1st nibble - flash 000000 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D4);	 							// 2nd nibble - flash 000001 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
}

void lcd_cursor(int x, int y){							// This function positions the cursor by converting a 20x4 coordinate to a DDRAM address line 1: 0..19, line 2: 64..83, line 3: 20..39, line 4: 84..103
	if(x<20){											// make sure x is no more than 39 - see p. 11
		x |= 0b10000000;								// set bit 7 to load DDRAM address - see p. 24
		if(y==1){
			x = x+64;									// if on line 2 add 0x40 to the x - see addresses p.11
		}
		if(y==2){										// if on line 3 you are using the last 20 addresses for line 1
			x = x+20;									// add 20 to the address value
		}
		if(y==3){										// if on line 4 you are using the last 20 addresses for line 2
			x = x+84;									// add 64 and 20 to your x value
		}
		LCD_PORT &= UN;									// clear data bus 
		LCD_PORT |= (EN + ((x >> 4) & 0x0f));			// high nibble - write x to the instruction register
		_delay_ms(1);										
		LCD_PORT &= UN;									// turn off pins
		_delay_ms(1);
		LCD_PORT |= (EN + (x & 0x0f));					// low nibble - write x
		_delay_ms(1);									
		LCD_PORT &= UN;									// turn off pins
		_delay_ms(1);
	}	
}
	
void lcd_data(const char d){							// loads byte to data register
	LCD_PORT &= UN;										// clear data bus
	LCD_PORT |= (EN + RS + ((d >> 4) & 0x0f));			// high nibble write data
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + RS + (d & 0x0f));					// low nibble write data
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
}		

void lcd_print(const char *s) {
	while(*s) {
		lcd_data(*s);
		s++;
	}
}		

void lcd_intprint(int i) {								// print 3 digit integers to ASCII
	unsigned int h,t,o;									// create variables for hundreds tens and ones
	h = (i)/100;										// divide integer by 100 to get hundreds digit
	if(h>0){											// if hundreds is not zero then
		lcd_data((h + 0x30));							// print 0x3h: ASCII code for h 
	}
	t = (i - (h*100))/10;								// subtract number of hundreds times 100 and divide by 10 to get tens 
	if((t>0)|(h>0)){									// if either tens or hundreds is greater than 0 then
		lcd_data(t + 0x30);								// print 0x3t: ASCII code for t 
	}
	o = (i - (h*100) - (t*10));							// subtract hundreds and tens to get ones and
	lcd_data(o + 0x30);									// print 0x3(o): ASCII code for o
}

int main(void){

	LCD_DDR |= 0b01111111;							// set all LCD pins to output
	BUT_DDR = 0x00;									// set all button pins to input
	BUT_PORT = 0xff;								// set pullup resistors on all button pins
	DDRB |= 0b00000001;								// set PB0 to output for relay
	
	lcd_init();										// initialize LCD
	
	while(1){
		int i = 0;
		lcd_cursor(0, 1);							// move the cursor to 0,1 beginning of second line 
		lcd_print("Welcome");						// print welcome 
		lcd_cursor(0, 2);							// move cursor to third line
		lcd_print("Awaiting Launch Command");		// print awaiting launch
			if(!(BUT_PIN & LAUNCH)){				// if the launch button is pressed
				lcd_clear();						// clear screen 
				lcd_cursor(0, 1);					// move cursor to line 2 
				lcd_print("Launch in");				// print launch in 	
				for(i=0, i<5000, i++){				// start a for loop polling 5000 cycles for ~5 seconds
					lcd_cursor(0, 2);				// move to third line 	
					lcd_intprint(5-(i/1000));		// hopefully print 5, 4, 3, 2, 1, 0 
					if(!(BUT_PIN & ABORT)){			// check if abort was pressed
						i = 5000;					// if so, set i to 5000 so we jump out of the loop
						lcd_clear();				// clear the screen 
						lcd_cursor(0, 1);			// move to second line 	
						lcd_print("Abort Launch");	// print abort launch 
						_delay_ms(5000);			// leave it there for 5 seconds 
						lcd_clear();				// clear the screen
					}								// otherwise
					if(i == 4999){					// when we get to 4999 	
						PORTB |= 0b00000001;		// trigger the relay 	
						lcd_clear();				// clear the screen 
						lcd_cursor(0, 1);			// go to second line 
						lcd_print("Launch Complete");	// print launch complete
						_delay_ms(1000);			// wait one second 
						PORTB &= 0b11111110;		// turn the relay off 	
						_delay_ms(4000);			// wait 4 seconds 	
						lcd_clear();				// clear the screen 	
					}								// until then 	
					_delay_ms(1);					// wait 1 millisecond and loop again
				}
			}
	}
	return(0);				
}