// I have not yet tested this code with the new bug fixes, but I am fairly confident it should work
// I changed the timing on the polling for loop in the main function (that part is tested)
// I rewrote the code for lcd_cursor and lcd_data functions to be independent of pin choice for the data bus 



// Launch controller
// coded by Phil Tirino

// 28 pin AVR Atmega328P
// 20x4 lcd screen on D bank
// PB3 PB4 PB5 SPI interface
// PD0 through PD6 used for LCD interface 
// PC0 through PC1 used for buttons

// Naming conventions
// All caps: register names, pin names and button names
// lower case underscore verbs: functions
// first word lower, cap second words; variables

#include <avr/io.h>                 // maps and defines names for SFR addresses       
#include <util/delay.h>				// prototypes _delay_ms and _delay_us

// pin names and port selection for LCD
// hook up RS to PD6, RW to PD5, EN to PD4, 
// hook up D4 to PD0, D5 to PD1, D6 to PD2, D7 to PD3 cannot change for lcd_data to work 

// hook up relay to PB0

// change these according to which pins the data bus is connected to
#define LCD_DDR 	DDRD			// defines data direction register for LCD
#define LCD_PORT 	PORTD			// defines port used for LCD
#define UN 			0b10000000		// unused pin, can be used as an output for other function
#define D7 			0b00001000		// 4 bit data bus 
#define D6 			0b00000100		// 4 bit data bus
#define D5 			0b00000010		// 4 bit data bus
#define D4 			0b00000001		// 4 bit data bus
#define EN 			0b00010000		// the enable pin triggers the data transfer
#define RW 			0b00100000		// read write: 0 for write
#define RS 			0b01000000		// register select: 0 for instruction register and 1 for data register

// pin names for buttons

#define BUT_DDR 	DDRC			// defines data direction register for buttons
#define BUT_PORT 	PORTC			// defines port for buttons
#define BUT_PIN 	PINC			// defines pin register for buttons
#define LAUNCH     	0b00000001		// for use in bit-mask for button
#define ABORT   	0b00000010		// for use in bit-mask for button 

#define ScrollDelay1 		5

void lcd_init(void) {
	LCD_PORT |= (EN + RS + RW + D7 + D6 + D5 + D4);		// flash all pins
	_delay_ms(20);										// follow reset sequence on p. 46 HD44780U datasheet
	LCD_PORT &= UN;							 			// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4 respectively
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(5);										// wait for more than 4.1 ms
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);										// wait more than 100 us
	LCD_PORT |= (EN + D5 + D4);							// flash 000011 for RS RW D7 D6 D5 D4
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);										// end of reset
	LCD_PORT |= (EN + D5);								// flash 000010 to initialize 4 bit data transfer mode: sequence on p. 42 HD44780U datasheet
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D5);								// 1st nibble - flash 000010 for FUNCTION: data length 4 - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D7);								// 2nd nibble - flash 001000 for FUNCTION: display lines 2 and 5x8 font- see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for DISPLAY: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D7 + D6);							// 2nd nibble - flash 001100 for DISPLAY: display on - see p. 24
	_delay_ms(1);											// add D5 to turn cursor on and D4 to turn blinking cursor position on
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for ENTRY: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D6 + D5);							// 2nd nibble - flash 000110 for ENTRY: auto increment and shift off - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= (EN + D4);								// 1st nibble - flash 000001 for CURSOR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D6);								// 2nd nibble - flash 000100 for CURSOR: cursor move shift right - see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;					 					// 1st nibble - flash 000000 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D4);	 							// 2nd nibble - flash 000001 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	LCD_PORT |= EN;										// 1st nibble - flash 000000 for HOME: see p. 24
	_delay_ms(2);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D5);	 							// 2nd nibble - flash 000010 for HOME: see p. 24
	_delay_ms(2);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
}

void lcd_clear(void){
	LCD_PORT |= EN;					 					// 1st nibble - flash 000000 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1); 
	LCD_PORT |= (EN + D4);	 							// 2nd nibble - flash 000001 for CLEAR: see p. 24
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
}

void lcd_cursor(int x, int y){							// This function positions the cursor by converting a 20x4 coordinate to a DDRAM address line 1: 0..19, line 2: 64..83, line 3: 20..39, line 4: 84..103
	int BUS = 0; 										// create a variable to hold data bus values
	if(x<20){											// make sure x is no more than 39 - see p. 11
		x |= 0b10000000;								// set bit 7 to load DDRAM address - see p. 24
		if(y==1){
			x = x+64;									// if on line 2 add 0x40 to the x - see addresses p.11
		}
		if(y==2){										// if on line 3 you are using the last 20 addresses for line 1
			x = x+20;									// add 20 to the address value
		}
		if(y==3){										// if on line 4 you are using the last 20 addresses for line 2
			x = x+84;									// add 64 and 20 to your x value
		}
		LCD_PORT &= UN;									// clear data bus 
		if((d >> 7) & 0b00000001){						// high nyble - send bit 7 to D7
			BUS = BUS + D7;								// if bit 7 is a 1 add D7 to the bus
		}
		if((d >> 6) & 0b00000001){						// high nyble - send bit 6 to D6
			BUS = BUS + D6;								// if bit 6 is a 1 add D6 to the bus
		}
		if((d >> 5) & 0b00000001){						// high nyble - send bit 5 to D5
			BUS = BUS + D5;								// if bit 5 is a 1 add D5 to the bus
		}
		if((d >> 4) & 0b00000001){						// high nyble - send bit 4 to D4
			BUS = BUS + D4;								// if bit 4 is a 1 add D4 to the bus
		}
		LCD_PORT |= (EN + BUS);							// high nibble - write x to the instruction register
		_delay_ms(1);										
		LCD_PORT &= UN;									// turn off pins
		_delay_ms(1);
		BUS = 0;										// clear BUS
		if((d >> 3) & 0b00000001){						// low nyble - send bit 3 to D7
			BUS = BUS + D7;								// if bit 3 is a 1 add D7 to bus 
		}
		if((d >> 2) & 0b00000001){						// low nyble - send bit 2 to D6 
			BUS = BUS + D6;								// if bit 2 is a 1 add D6 to bus 
		}
		if((d >> 1) & 0b00000001){						// low nyble - send bit 1 to D5 
			BUS = BUS + D5;								// if bit 1 is a 1 add D4 to bus 
		}
		if((d) & 0b00000001){							// low nyble - send bit 0 to D4 
			BUS = BUS + D4;								// if bit 0 is a 1 add D4 to bus
		}
		LCD_PORT |= (EN + BUS);							// low nibble - write x
		_delay_ms(1);									
		LCD_PORT &= UN;									// turn off pins
		_delay_ms(1);
		BUS = 0;
	}	
}

void lcd_data(const char d){							// loads byte to data register
	int BUS = 0; 										// create a variable to hold data bus values
	LCD_PORT &= UN;										// clear data busd
	if((d >> 7) & 0b00000001){							// high nyble - send bit 7 to D7
		BUS = BUS + D7;									// if bit 7 is a 1 add D7 to the bus
	}
	if((d >> 6) & 0b00000001){							// high nyble - send bit 6 to D6
		BUS = BUS + D6;									// if bit 6 is a 1 add D6 to the bus
	}
	if((d >> 5) & 0b00000001){							// high nyble - send bit 5 to D5
		BUS = BUS + D5;									// if bit 5 is a 1 add D5 to the bus
	}
	if((d >> 4) & 0b00000001){							// high nyble - send bit 4 to D4
		BUS = BUS + D4;									// if bit 4 is a 1 add D4 to the bus
	}
	LCD_PORT |= (EN + RS + BUS);						// high nibble write data
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	BUS = 0;											// clear BUS
	if((d >> 3) & 0b00000001){							// low nyble - send bit 3 to D7
		BUS = BUS + D7;									// if bit 3 is a 1 add D7 to bus 
	}
	if((d >> 2) & 0b00000001){							// low nyble - send bit 2 to D6 
		BUS = BUS + D6;									// if bit 2 is a 1 add D6 to bus 
	}
	if((d >> 1) & 0b00000001){							// low nyble - send bit 1 to D5 
		BUS = BUS + D5;									// if bit 1 is a 1 add D4 to bus 
	}
	if((d) & 0b00000001){								// low nyble - send bit 0 to D4 
		BUS = BUS + D4;									// if bit 0 is a 1 add D4 to bus
	}
	LCD_PORT |= (EN + RS + BUS);			   			// low nibble write data
	_delay_ms(1);										
	LCD_PORT &= UN;										// turn off pins
	_delay_ms(1);
	BUS = 0;
}	

void lcd_print(const char *s) {
	while(*s) {
		lcd_data(*s);
		s++;
	}
}		

void lcd_intprint(int i) {								// print 3 digit integers to ASCII
	unsigned int h,t,o;									// create variables for hundreds tens and ones
	h = (i)/100;										// divide integer by 100 to get hundreds digit
	if(h>0){											// if hundreds is not zero then
		lcd_data((h + 0x30));							// print 0x3h: ASCII code for h 
	}
	t = (i - (h*100))/10;								// subtract number of hundreds times 100 and divide by 10 to get tens 
	if((t>0)|(h>0)){									// if either tens or hundreds is greater than 0 then
		lcd_data(t + 0x30);								// print 0x3t: ASCII code for t 
	}
	o = (i - (h*100) - (t*10));							// subtract hundreds and tens to get ones and
	lcd_data(o + 0x30);									// print 0x3(o): ASCII code for o
}

int main(void){

	LCD_DDR |= 0b01111111;							// set all LCD pins to output
	BUT_DDR = 0x00;									// set all button pins to input
	BUT_PORT = 0xff;								// set pullup resistors on all button pins
	DDRB |= 0b00000001;								// set PB0 to output for relay
	PORTB |= 0b00000001;							// set output to high
	
	lcd_init();										// initialize LCD
	
	while(1){
		int i = 0;
		lcd_cursor(0, 0);							// move the cursor to 0,0 beginning of first line 
		lcd_print("Welcome, Awaiting");				// print welcome, awaiting 
		lcd_cursor(0, 1);							// move cursor to second line
		lcd_print("Launch Command...");				// print launch command
		lcd_cursor(0, 2);							// move to third line 
		lcd_print("Left Btn   Right Btn");			// print left btn right btn
		lcd_cursor(0, 3);							// move to fourth line 
		lcd_print("LAUNCH     ABORT");				// print LAUNCH ABORT
			if(!(BUT_PIN & LAUNCH)){				// if the launch button is pressed
				lcd_clear();						// clear screen 
				lcd_cursor(0, 1);					// move cursor to line 2 
				lcd_print("Launch in");				// print launch in 	
				for(i=0; i<625; i++){				// start a for loop polling 5000 cycles for ~5 seconds
					lcd_cursor(0, 2);				// move to third line 	***has 4ms delay
					lcd_intprint(5-(i/125));		// hopefully print 5, 4, 3, 2, 1, 0 ***has 4ms delay
					if(!(BUT_PIN & ABORT)){			// check if abort was pressed
						i = 625;					// if so, set i to 625 so we jump out of the loop
						lcd_clear();				// clear the screen 
						lcd_cursor(0, 1);			// move to second line 	
						lcd_print("Abort Launch");	// print abort launch 
						_delay_ms(5000);			// leave it there for 5 seconds 
						lcd_clear();				// clear the screen
					}								// otherwise
					if(i == 624){					// when we get to 624	
						PORTB &= 0b11111110;		// trigger the relay 	
						lcd_clear();				// clear the screen 
						lcd_cursor(0, 1);			// go to second line 
						lcd_print("Launch Complete");	// print launch complete
						_delay_ms(1500);			// wait one second 
						PORTB |= 0b00000001;;		// turn the relay off 	
						_delay_ms(4000);			// wait 4 seconds 	
						lcd_clear();				// clear the screen 	
					}								// until then 	
					_delay_ms(1);					// wait 1 millisecond and loop again
				}
			}
	}
	return(0);				
}